
from .ldata import BLACK_SQUARES
from pychess.Utils.const import ASEAN_VARIANTS, WHITE, BLACK, KNIGHT, BISHOP, ROOK, QUEEN, PAWN


def testFifty(board):
    if board.fifty >= 100:
        return True
    return False


drawSet = set(((0, 0, 0, 0, 0, 0, 0, 0),  # KK
               (0, 1, 0, 0, 0, 0, 0, 0),  # KBK
               (1, 0, 0, 0, 0, 0, 0, 0),  # KNK
               (0, 0, 0, 0, 0, 1, 0, 0),  # KKB
               (0, 0, 0, 0, 1, 0, 0, 0),  # KNK
               (1, 0, 0, 0, 0, 1, 0, 0),  # KNKB
               (0, 1, 0, 0, 1, 0, 0, 0),  # KBKN
               ))

# Contains not 100% sure ones
drawSet2 = set(((2, 0, 0, 0, 0, 0, 0, 0),  # KNNK
                (0, 0, 0, 0, 2, 0, 0, 0),  # KKNN
                (2, 0, 0, 0, 1, 0, 0, 0),  # KNNKN
                (1, 0, 0, 0, 2, 0, 0, 0),  # KNKNN
                (2, 0, 0, 0, 0, 1, 0, 0),  # KNNKB
                (0, 1, 0, 0, 2, 0, 0, 0),  # KBKNN
                (2, 0, 0, 0, 0, 0, 1, 0),  # KNNKR
                (0, 0, 1, 0, 2, 0, 0, 0)   # KRKNN
                ))


def testMaterial(board):
    """ Tests if no players are able to win the game from the current
        position """

    whitePieceCount = board.pieceCount[WHITE]
    blackPieceCount = board.pieceCount[BLACK]

    if whitePieceCount[PAWN] or blackPieceCount[PAWN]:
        return False

    if whitePieceCount[QUEEN] or blackPieceCount[QUEEN]:
        return False

    wn = whitePieceCount[KNIGHT]
    wb = whitePieceCount[BISHOP]
    wr = whitePieceCount[ROOK]
    bn = blackPieceCount[KNIGHT]
    bb = blackPieceCount[BISHOP]
    br = blackPieceCount[ROOK]

    if board.variant in ASEAN_VARIANTS and (wb or bb):
        return False

    if (wn, wb, wr, 0, bn, bb, br, 0) in drawSet:
        return True

    # Tests KBKB. Draw if bishops are of same color
    if not wn + wr + bn + br and wb == 1 and bb == 1:
        if board.boards[WHITE][BISHOP] & BLACK_SQUARES != \
           board.boards[BLACK][BISHOP] & BLACK_SQUARES:
            return True


def testPlayerMatingMaterial(board, color):
    """ Tests if given color has enough material to mate on board """

    pieceCount = board.pieceCount[color]

    if pieceCount[PAWN] or pieceCount[QUEEN] or pieceCount[ROOK] \
       or (pieceCount[KNIGHT] + pieceCount[BISHOP] > 1):
        return True
    return False

# This could be expanded by the fruit kpk draw function, which can test if a
# certain king verus king and pawn posistion is winable.


def test(board):
    """ Test if the position is drawn. Two-fold repetitions are counted. """
    return board.repetitionCount(draw_threshold=2) > 1 or \
        testFifty(board) or \
        testMaterial(board)



from pychess.Utils.const import EUROSHOGICHESS, VARIANTS_OTHER_NONSTANDARD, \
    A8, B8, C8, D8, E8, F8, G8, H8, \
    A7, B7, C7, D7, E7, F7, G7, H7, \
    A6, B6, C6, D6, E6, F6, G6, H6, \
    A1, B1, C1, D1, E1, F1, G1, H1, \
    A2, B2, C2, D2, E2, F2, G2, H2, \
    A3, B3, C3, D3, E3, F3, G3, H3


from pychess.Utils.Board import Board

EUROSHOGISTART = "1nbqkqn1/1r4b1/pppppppp/8/8/PPPPPPPP/1B4R1/1NQKQBN1 w - - 0 1"


class EuroShogiBoard(Board):
    variant = EUROSHOGICHESS
    __desc__ = _("EuroShogi: http://en.wikipedia.org/wiki/EuroShogi")
    name = _("EuroShogi")
    cecp_name = "euroshogi"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    PROMOTION_ZONE = ((A8, B8, C8, D8, E8, F8, G8, H8,
                       A7, B7, C7, D7, E7, F7, G7, H7,
                       A6, B6, C6, D6, E6, F6, G6, H6),
                      (A1, B1, C1, D1, E1, F1, G1, H1,
                       A2, B2, C2, D2, E2, F2, G2, H2,
                       A3, B3, C3, D3, E3, F3, G3, H3))

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=EUROSHOGISTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import datetime

from gi.repository import GObject
from .BoardManager import names, months, dates

from pychess.ic import GAME_TYPES_BY_SHORT_FICS_NAME, BLKCMD_STORED, \
    BLKCMD_HISTORY, BLKCMD_JOURNAL
from pychess.ic.FICSObjects import FICSAdjournedGame, FICSHistoryGame, FICSJournalGame

from pychess.Utils.const import WON_ADJUDICATION, DRAW_AGREE, WON_DISCONNECTION, WON_CALLFLAG, \
    WON_MATE, DRAW_INSUFFICIENT, DRAW_REPETITION, WON_RESIGN, DRAW_STALEMATE, \
    DRAW_BLACKINSUFFICIENTANDWHITETIME, WON_NOMATERIAL, DRAW_50MOVES, WHITEWON, DRAW, \
    BLACK, WHITE, BLACKWON, reprResult, ADJOURNED

from pychess.System.Log import log

reasons_dict = {
    "Adj": WON_ADJUDICATION,
    "Agr": DRAW_AGREE,
    "Dis": WON_DISCONNECTION,
    "Fla": WON_CALLFLAG,
    "Mat": WON_MATE,
    "NM": DRAW_INSUFFICIENT,
    "Rep": DRAW_REPETITION,
    "Res": WON_RESIGN,
    "Sta": DRAW_STALEMATE,
    "TM":
    DRAW_BLACKINSUFFICIENTANDWHITETIME,  # DRAW_WHITEINSUFFICIENTANDBLACKTIME
    "WLM": WON_NOMATERIAL,
    "WNM": WON_NOMATERIAL,
    "50": DRAW_50MOVES
}

reasons = "(%s)" % "|".join(reasons_dict.keys())
ratings = r"([0-9\ \-\+]{1,4}[P E]?|UNR)"


class AdjournManager(GObject.GObject):

    __gsignals__ = {
        'adjournedGameAdded': (GObject.SignalFlags.RUN_FIRST, None,
                               (object, )),
        'onAdjournmentsList': (GObject.SignalFlags.RUN_FIRST, None,
                               (object, )),
        'historyGameAdded': (GObject.SignalFlags.RUN_FIRST, None, (object, )),
        'onHistoryList': (GObject.SignalFlags.RUN_FIRST, None, (object, )),
        'journalGameAdded': (GObject.SignalFlags.RUN_FIRST, None, (object, )),
        'onJournalList': (GObject.SignalFlags.RUN_FIRST, None, (object, )),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)

        self.connection = connection

        self.connection.expect_line(self.__onStoredResponseNO,
                                    r"%s has no adjourned games\." % names)

        self.connection.expect_line(self.__onHistoryResponseNO,
                                    r"%s has no history games\." % names)

        self.connection.expect_line(
            self.__onJournalResponseNO,
            r"(%s has no journal entries\.)|(That journal is private.)" % names)

        self.connection.expect_fromABplus(
            self.__onStoredResponseYES, "Stored games for %s:" % names,
            r"\s*C Opponent\s+On Type\s+Str\s+M\s+ECO\s+Date",
            r"\s*\d+: (B|W) %s\s+(Y|N) \[([a-z ]{3})\s+(\d+)\s+(\d+)\]\s+(\d+)-(\d+)\s+(W|B)(\d+)\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s"
            % (names, dates))

        self.connection.expect_fromABplus(
            self.__onHistoryResponseYES, "History for %s:" % names,
            r"\s*Opponent\s+Type\s+ECO\s+End\s+Date",
            r"\s*(\d+): (-|\+|=)\s+(\d+)\s+(W|B)\s+(\d+) %s\s+\[([a-z ]{3})\s*(\d+)\s+(\d+)\]\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s\s+%s"
            % (names, reasons, dates))

        self.connection.expect_fromABplus(
            self.__onJournalResponseYES, "Journal for %s:" % names,
            r"\s*White\s+Rating\s+Black\s+Rating\s+Type\s+ECO\s+End\s+Result",
            r"\s*%%(\d+): %s\s+%s\s+%s\s+%s\s+\[([a-z ]{3})\s+(\d+)\s+(\d+)\]\s+(---|\?\?\?|\*\*\*|[A-Z]\d+)\s+%s\s+(\*|1/2-1/2|1-0|0-1)"
            % (names, ratings, names, ratings, reasons))

        self.connection.expect_line(self.__onAdjournedGameResigned,
                                    r"You have resigned the game\.")

        self.connection.bm.connect("curGameEnded", self.__onCurGameEnded)

        self.queryAdjournments()
        self.queryHistory()
        self.queryJournal()

        # TODO: Connect to {Game 67 (MAd vs. Sandstrom) Game adjourned by mutual agreement} *
        # TODO: Connect to adjourned game as adjudicated

    def __onStoredResponseYES(self, matchlist):
        # Stored games for User:
        #     C Opponent     On Type          Str  M    ECO Date
        #  1: W TheDane       N [ br  2  12]  0-0  B2   ??? Sun Nov 23,  6:14 CST 1997
        #  2: W PyChess       Y [psu  2  12] 39-39 W3   C20 Sun Jan 11, 17:40 ??? 2009
        #  3: B cjavad        N [ wr  2   2] 31-31 W18  --- Wed Dec 23, 06:58 PST 2009
        self.connection.stored_owner = matchlist[0].groups()[0]
        adjournments = []
        for match in matchlist[2:]:
            our_color = match.groups()[0]
            opponent_name, opponent_online = match.groups()[1:3]
            game_type = match.groups()[3]
            minutes, gain = match.groups()[4:6]
            str_white, str_black = match.groups()[6:8]
            next_color = match.groups()[8]
            move_num = match.groups()[9]
            week, month, day, hour, minute, timezone, year = match.groups()[11:18]
            gametime = datetime.datetime(
                int(year), months.index(month) + 1, int(day), int(hour),
                int(minute))
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            our_color = our_color == "B" and BLACK or WHITE
            minutes = int(minutes)
            gain = int(gain)
            length = (int(move_num) - 1) * 2
            if next_color == "B":
                length += 1

            user = self.connection.players.get(self.connection.stored_owner)
            opponent = self.connection.players.get(opponent_name)
            wplayer, bplayer = (user, opponent) if our_color == WHITE else (opponent, user)
            game = FICSAdjournedGame(wplayer,
                                     bplayer,
                                     game_type=gametype,
                                     rated=rated,
                                     our_color=our_color,
                                     length=length,
                                     time=gametime,
                                     minutes=minutes,
                                     inc=gain,
                                     private=private)
            if game.opponent.adjournment is False:
                game.opponent.adjournment = True

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("adjournedGameAdded", game)
            adjournments.append(game)

        self.emit("onAdjournmentsList", adjournments)

    __onStoredResponseYES.BLKCMD = BLKCMD_STORED

    def __onHistoryResponseYES(self, matchlist):
        # History for User:
        # Opponent      Type         ECO End Date
        # 66: - 1735 B    0 GuestHKZX     [ bu  3   0] B23 Res Sun Dec  6, 15:50 EST 2015
        # 67: - 1703 B    0 GuestQWML     [ lu  1   0] B07 Fla Sun Dec  6, 15:53 EST 2015
        history = []
        self.connection.history_owner = matchlist[0].groups()[0]
        for match in matchlist[2:]:
            # print(match.groups())
            history_no = match.groups()[0]
            result = match.groups()[1]
            owner_rating = match.groups()[2]
            owner_color = match.groups()[3]
            opp_rating = match.groups()[4]
            if result == "+":
                result = WHITEWON if owner_color == "W" else BLACKWON
            elif result == "-":
                result = WHITEWON if owner_color == "B" else BLACKWON
            else:
                result = DRAW
            opponent_name = match.groups()[5]
            if owner_color == "W":
                white = self.connection.history_owner
                black = opponent_name
                wrating = owner_rating
                brating = opp_rating
            else:
                white = opponent_name
                black = self.connection.history_owner
                brating = owner_rating
                wrating = opp_rating
            game_type = match.groups()[6]
            minutes, gain = match.groups()[7:9]
            reason = reasons_dict[match.groups()[10]]
            week, month, day, hour, minute, timezone, year = match.groups()[11:18]
            gametime = datetime.datetime(
                int(year), months.index(month) + 1, int(day), int(hour),
                int(minute))
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            owner_color = owner_color == "B" and BLACK or WHITE
            minutes = int(minutes)
            gain = int(gain)

            wplayer = self.connection.players.get(white)
            bplayer = self.connection.players.get(black)
            game = FICSHistoryGame(wplayer,
                                   bplayer,
                                   game_type=gametype,
                                   rated=rated,
                                   minutes=minutes,
                                   inc=gain,
                                   private=private,
                                   wrating=wrating,
                                   brating=brating,
                                   time=gametime,
                                   reason=reason,
                                   history_no=history_no,
                                   result=result)

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("historyGameAdded", game)
            history.append(game)

        self.emit("onHistoryList", history)

    __onHistoryResponseYES.BLKCMD = BLKCMD_HISTORY

    def __onJournalResponseYES(self, matchlist):
        # Journal for User:
        #     White         Rating  Black         Rating  Type         ECO End Result
        # %01: tentacle      2291    larsa         2050    [ lr  1   2] D35 Rep 1/2-1/2
        # %02: larsa         2045    tentacle      2296    [ lr  1   2] A46 Res 0-1
        journal = []
        self.connection.journal_owner = matchlist[0].groups()[0]
        for match in matchlist[2:]:
            # print(match.groups())
            journal_no = match.groups()[0]
            result = match.groups()[10]
            result = reprResult.index(result)
            white = match.groups()[1]
            wrating = match.groups()[2]
            black = match.groups()[3]
            brating = match.groups()[4]
            game_type = match.groups()[5]
            minutes, gain = match.groups()[6:8]
            reason = reasons_dict[match.groups()[9]]
            private = game_type[0] == "p"
            rated = game_type[2] == "r"
            gametype = GAME_TYPES_BY_SHORT_FICS_NAME[game_type[1]]
            minutes = int(minutes)
            gain = int(gain)

            wplayer = self.connection.players.get(white)
            bplayer = self.connection.players.get(black)
            game = FICSJournalGame(wplayer,
                                   bplayer,
                                   game_type=gametype,
                                   rated=rated,
                                   minutes=minutes,
                                   inc=gain,
                                   private=private,
                                   wrating=wrating,
                                   brating=brating,
                                   reason=reason,
                                   journal_no=journal_no,
                                   result=result)

            if game not in self.connection.games:
                game = self.connection.games.get(game, emit=False)
                self.emit("journalGameAdded", game)
            journal.append(game)

        self.emit("onJournalList", journal)

    __onJournalResponseYES.BLKCMD = BLKCMD_JOURNAL

    def __onStoredResponseNO(self, match):
        self.connection.stored_owner = match.groups()[0]
        self.emit("onAdjournmentsList", [])

    __onStoredResponseNO.BLKCMD = BLKCMD_STORED

    def __onHistoryResponseNO(self, match):
        self.connection.history_owner = match.groups()[0]
        self.emit("onHistoryList", [])

    __onHistoryResponseNO.BLKCMD = BLKCMD_HISTORY

    def __onJournalResponseNO(self, match):
        self.connection.journal_owner = match.groups()[0]
        self.emit("onJournalList", [])

    __onJournalResponseNO.BLKCMD = BLKCMD_JOURNAL

    def __onAdjournedGameResigned(self, match):
        self.queryAdjournments()

    def __onCurGameEnded(self, bm, game):
        if game.result == ADJOURNED:
            self.queryAdjournments()
        elif game.result in (DRAW, WHITEWON, BLACKWON):
            self.queryHistory()

    def queryAdjournments(self, owner=None):
        if owner is None:
            self.connection.client.run_command("stored")
        else:
            self.connection.client.run_command("stored %s" % owner)

    def queryHistory(self, owner=None):
        if owner is None:
            self.connection.client.run_command("history")
        else:
            self.connection.client.run_command("history %s" % owner)

    def queryJournal(self, owner=None):
        if owner is None:
            self.connection.client.run_command("journal")
        else:
            self.connection.client.run_command("journal %s" % owner)

    def queryMoves(self, game):
        if isinstance(game, FICSHistoryGame):
            self.connection.client.run_command("smoves %s %s" % (
                self.connection.history_owner, game.history_no))
        elif isinstance(game, FICSJournalGame):
            self.connection.client.run_command("smoves %s %%%s" % (
                self.connection.journal_owner, game.journal_no))
        else:
            self.connection.client.run_command("smoves %s %s" % (
                self.connection.stored_owner, game.opponent.name))

    def examine(self, game):
        game.board = None
        self.connection.archived_examine = game
        if isinstance(game, FICSAdjournedGame):
            self.connection.client.run_command("examine %s %s" % (
                self.connection.stored_owner, game.opponent.name))
        elif isinstance(game, FICSHistoryGame):
            self.connection.client.run_command("examine %s %s" % (
                self.connection.history_owner, game.history_no))
        elif isinstance(game, FICSJournalGame):
            self.connection.client.run_command("examine %s %%%s" % (
                self.connection.journal_owner, game.journal_no))

    def challenge(self, playerName):
        self.connection.client.run_command("match %s" % playerName)

    def resign(self, game):
        """ This is (and draw and abort) are possible even when one's
            opponent is not logged on """
        if not game.opponent.adjournment:
            log.warning("AdjournManager.resign: no adjourned game vs %s" %
                        game.opponent)
            return
        log.info("AdjournManager.resign: resigning adjourned game=%s" % game)
        self.connection.client.run_command("resign %s" % game.opponent.name)

    def draw(self, game):
        if not game.opponent.adjournment:
            log.warning("AdjournManager.draw: no adjourned game vs %s" %
                        game.opponent)
            return
        log.info("AdjournManager.draw: offering sdraw for adjourned game=%s" %
                 game)
        self.connection.client.run_command("sdraw %s" % game.opponent.name)

    def abort(self, game):
        if not game.opponent.adjournment:
            log.warning("AdjournManager.abort: no adjourned game vs %s" %
                        game.opponent)
            return
        log.info("AdjournManager.abort: offering sabort for adjourned game=%s"
                 % game)
        self.connection.client.run_command("sabort %s" % game.opponent.name)

    def resume(self, game):
        if not game.opponent.adjournment:
            log.warning("AdjournManager.resume: no adjourned game vs %s" %
                        game.opponent)
            return
        log.info("AdjournManager.resume: offering resume for adjourned game=%s"
                 % game)
        self.connection.client.run_command("match %s" % game.opponent.name)

    # (a)  Users who have more than 15 stored games are restricted from starting new
    # games.  If this situation happens to you, review your stored games and see
    # which ones might be eligible for adjudication (see "help adjudication").



# -*- coding: UTF-8 -*-

from glob import glob
from os import listdir
from os.path import isdir, isfile
import os
import site
import sys
import subprocess

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path = [os.path.join(this_dir, "lib")] + sys.path

from pychess.Savers.pgn import PGNFile
from pychess.System.protoopen import protoopen


msi = False
if sys.argv[-1] == "bdist_msi":
    try:
        from cx_Freeze import setup, Executable
        from cx_Freeze.windist import bdist_msi
        msi = True
    except ImportError:
        print("ERROR: can't import cx_Freeze!")
        sys.exit(1)
else:
    from distutils.core import setup

if sys.version_info < (3, 5, 0):
    print('ERROR: PyChess requires Python >= 3.5.0')
    sys.exit(1)

if sys.platform == "win32":
    try:
        from gi.repository import Gtk
        print("Gtk verion is %s.%s.%s", (Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION, Gtk.MICRO_VERSION))
    except ImportError:
        print('ERROR: PyChess in Windows Platform requires to install PyGObject.')
        print('Installing from http://sourceforge.net/projects/pygobjectwin32')
        sys.exit(1)

from imp import load_module, find_module
pychess = load_module("pychess", *find_module("pychess", ["lib"]))

VERSION = pychess.VERSION

NAME = "pychess"

# We have to subclass register command because
# PyChess from another author already exist on pypi.

from distutils.command.register import register


class RegisterCommand(register):
    def run(self):
        self.distribution.metadata.name = "PyChess-%s" % pychess.VERSION_NAME
        register.run(self)


DESC = "Chess client"

LONG_DESC = """PyChess is a chess client for playing and analyzing chess games. It is
intended to be usable both for those totally new to chess as well as
advanced users who want to use a computer to further enhance their play.

PyChess has a builtin python chess engine and auto-detects most
popular chess engines (Stockfish, Rybka, Houdini, Shredder, GNU Chess,
Crafty, Fruit, and many more). These engines are available as opponents,
and are used to provide hints and analysis. PyChess also shows analysis
from opening books and Gaviota end-game tablebases.

When you get sick of playing computer players you can login to FICS (the
Free Internet Chess Server) and play against people all over the world.
PyChess has a built-in Timeseal client, so you won't lose clock time during
a game due to lag. PyChess also has pre-move support, which means you can
make (or start making) a move before your opponent has made their move.

PyChess has many other features including:
- CECP and UCI chess engine support with customizable engine configurations
- Polyglot opening book support
- Hint and Spy move arrows
- Hint, Score, and Annotation panels
- Play and analyze games in separate game tabs
- 18 chess variants including Chess960, Suicide, Crazyhouse, Shuffle, Losers, Piece Odds, and Atomic
- Reads and writes PGN, EPD and FEN chess file formats
- Undo and pause chess games
- Move animation in games
- Drag and drop chess files
- Optional game move and event sounds
- Chess piece themes with 40 built-in piece themes
- Legal move highlighting
- Direct copy+paste pgn game input via Enter Game Notation open-game dialog
- Internationalised text and Figurine Algebraic Notation (FAN) support
- Translated into 38 languages (languages with +5% strings translated)
- Easy to use and intuitive look and feel"""

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: X11 Applications :: GTK',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Games/Entertainment :: Board Games']

os.chdir(os.path.abspath(os.path.dirname(__file__)))

# save
stderr = sys.stderr
stdout = sys.stdout

if not isfile("eco.db"):
    with open("pgn2ecodb.py") as fh:
        exec(fh.read())

if not isfile(os.path.abspath("pieces/Pychess.png")):
    with open("create_theme_preview.py") as fh:
        exec(fh.read())

# restore
sys.stderr = stderr
sys.stdout = stdout

DATA_FILES = [("share/pychess",
               ["README.md", "AUTHORS", "ARTISTS", "DOCUMENTERS",
                "LICENSE", "TRANSLATORS", "pychess_book.bin", "eco.db"])]

# UI
DATA_FILES += [("share/pychess/glade", glob('glade/*.glade'))]
DATA_FILES += [("share/pychess/glade", ['glade/background.jpg'])]
DATA_FILES += [("share/pychess/glade", glob('glade/*.png'))]
DATA_FILES += [("share/pychess/glade/16x16", glob('glade/16x16/*.png'))]
DATA_FILES += [("share/pychess/glade/48x48", glob('glade/48x48/*.png'))]
DATA_FILES += [("share/pychess/glade", glob('glade/*.svg'))]
DATA_FILES += [("share/pychess/flags", glob('flags/*.png'))]
DATA_FILES += [("share/pychess/boards", glob('boards/*.png'))]

# Data
DATA_FILES += [('share/mime/packages', ['pychess.xml'])]
DATA_FILES += [('share/metainfo', ['pychess.metainfo.xml'])]
DATA_FILES += [('share/applications', ['pychess.desktop'])]
DATA_FILES += [('share/icons/hicolor/scalable/apps', ['pychess.svg'])]
DATA_FILES += [('share/menu', ['menu/pychess'])]
DATA_FILES += [('share/pixmaps', ['pychess.svg', 'pychess.xmp'])]
if sys.platform == "win32":
    DATA_FILES += [("share/pychess/sounds", glob('sounds/*.wav'))]
    DATA_FILES += [("share/pychess/engines", glob('engines/*.*'))]
else:
    DATA_FILES += [("share/pychess/sounds", glob('sounds/*.ogg'))]
DATA_FILES += [('share/icons/hicolor/24x24/apps', ['pychess.png'])]
DATA_FILES += [('share/gtksourceview-3.0/language-specs', ['gtksourceview-3.0/language-specs/pgn.lang'])]

# Piece sets
DATA_FILES += [("share/pychess/pieces", glob('pieces/*.png'))]
DATA_FILES += [("share/pychess/pieces/ttf", glob('pieces/ttf/*.ttf'))]

# Lectures, puzzles, lessons
for filename in glob('learn/puzzles/*.pgn'):
    chessfile = PGNFile(protoopen(filename))
    chessfile.init_tag_database()

for filename in glob('learn/lessons/*.pgn'):
    chessfile = PGNFile(protoopen(filename))
    chessfile.init_tag_database()

DATA_FILES += [("share/pychess/learn/puzzles", glob('learn/puzzles/*.olv'))]
DATA_FILES += [("share/pychess/learn/puzzles", glob('learn/puzzles/*.pgn'))]
DATA_FILES += [("share/pychess/learn/puzzles", glob('learn/puzzles/*.sqlite'))]
DATA_FILES += [("share/pychess/learn/lessons", glob('learn/lessons/*.pgn'))]
DATA_FILES += [("share/pychess/learn/lessons", glob('learn/lessons/*.sqlite'))]
DATA_FILES += [("share/pychess/learn/lectures", glob('learn/lectures/*.txt'))]

for dir in [d for d in listdir('pieces') if isdir(os.path.join('pieces', d)) and d != 'ttf']:
    DATA_FILES += [("share/pychess/pieces/" + dir, glob('pieces/' + dir + '/*.svg'))]

# Manpages
DATA_FILES += [('share/man/man1', ['manpages/pychess.1.gz'])]

# Language
pofile = "LC_MESSAGES/pychess"
if sys.platform == "win32":
    argv0_path = os.path.dirname(os.path.abspath(sys.executable))
    if pychess.MSYS2:
        major, minor, micro, releaselevel, serial = sys.version_info
        msgfmt_path = argv0_path + "/../lib/python%s.%s/tools/i18n/" % (major, minor)
    else:
        msgfmt_path = argv0_path + "/tools/i18n/"
    msgfmt = "%s %smsgfmt.py" % (os.path.abspath(sys.executable), msgfmt_path)
else:
    msgfmt = "msgfmt"

pychess_langs = []
for dir in [d for d in listdir("lang") if isdir("lang/" + d) and d != "en"]:
    if sys.platform == "win32":
        command = "%s lang/%s/%s.po" % (msgfmt, dir, pofile)
    else:
        command = "%s lang/%s/%s.po -o lang/%s/%s.mo" % (msgfmt, dir, pofile, dir, pofile)
    subprocess.call(command.split())
    DATA_FILES += [("share/locale/" + dir + "/LC_MESSAGES", ["lang/" + dir + "/" + pofile + ".mo"])]
    pychess_langs.append(dir)

PACKAGES = []

if msi:
    if pychess.MSYS2:
        gtk_data_path = sys.prefix
        gtk_exec_path = os.path.join(sys.prefix, "bin")
        lang_path = os.path.join(sys.prefix, "share", "locale")
    else:
        # Get the site-package folder, not everybody will install
        # Python into C:\PythonXX
        site_dir = site.getsitepackages()[1]
        gtk_data_path = os.path.join(site_dir, "gnome")
        gtk_exec_path = os.path.join(site_dir, "gnome")
        lang_path = os.path.join(site_dir, "gnome", "share", "locale")

    # gtk3.0 .mo files
    gtk_mo = [f + "/LC_MESSAGES/gtk30.mo" for f in os.listdir(lang_path) if f in pychess_langs]

    # Collect the list of missing dll when cx_freeze builds the app
    gtk_exec = ['libgtksourceview-3.0-1.dll',
                'libjpeg-8.dll',
                'librsvg-2-2.dll',
                ]

    # We need to add all the libraries too (for themes, etc..)
    gtk_data = ['etc',
                'lib/gdk-pixbuf-2.0',
                'lib/girepository-1.0',
                'share/icons/adwaita/icon-theme.cache',
                'share/icons/adwaita/index.theme',
                'share/icons/adwaita/16x16',
                'share/icons/adwaita/24x24',
                'share/icons/adwaita/48x48',
                'share/glib-2.0']

    # Create the list of includes as cx_freeze likes
    include_files = []
    for mo in gtk_mo:
        mofile = os.path.join(lang_path, mo)
        if os.path.isfile(mofile):
            include_files.append((mofile, "share/locale/" + mo))

    for dll in gtk_exec:
        include_files.append((os.path.join(gtk_exec_path, dll), dll))

    # Let's add gtk data
    for lib in gtk_data:
        include_files.append((os.path.join(gtk_data_path, lib), lib))

    base = None
    # Lets not open the console while running the app
    if sys.platform == "win32":
        base = "Win32GUI"

    executables = [Executable("pychess",
                              base=base,
                              icon="pychess.ico",
                              shortcutName="PyChess",
                              shortcutDir="DesktopFolder"),
                   Executable(script="lib/__main__.py",
                              targetName="pychess-engine.exe",
                              base=base)]

    bdist_msi_options = {
        "upgrade_code": "{5167584f-c196-428f-be40-4c861025e90a}",
        "add_to_path": False}

    perspectives = ["pychess.perspectives"]
    for persp in ("welcome", "games", "fics", "database", "learn"):
        perspectives.append("pychess.perspectives.%s" % persp)

    build_exe_options = {
        "path": sys.path + ["lib"],
        "includes": ["gi"],
        "packages": ["asyncio", "gi", "sqlalchemy.dialects.sqlite", "sqlalchemy.sql.default_comparator", "pexpect", "pychess"] + perspectives,
        "include_files": include_files}
    if pychess.MSYS2:
        build_exe_options["excludes"] = ["tkinter"]
    else:
        build_exe_options["include_msvcr"] = True

else:
    PACKAGES = ["pychess", "pychess.gfx", "pychess.ic", "pychess.ic.managers",
                "pychess.Players", "pychess.Savers", "pychess.System",
                "pychess.Utils", "pychess.Utils.lutils", "pychess.Variants",
                "pychess.Database", "pychess.widgets", "pychess.widgets.pydock",
                "pychess.perspectives", "pychess.perspectives.welcome",
                "pychess.perspectives.games", "pychess.perspectives.fics",
                "pychess.perspectives.database", "pychess.perspectives.learn",
                "pychess.external"]

    build_exe_options = {}
    bdist_msi_options = {}
    executables = {}

setup(
    cmdclass={"register": RegisterCommand},
    name=NAME,
    version=VERSION,
    author='Pychess team',
    author_email='pychess-people@googlegroups.com',
    maintainer='Thomas Dybdahl Ahle',
    classifiers=CLASSIFIERS,
    keywords='python gtk chess xboard gnuchess game pgn epd board linux',
    description=DESC,
    long_description=LONG_DESC,
    license='GPL3',
    url='http://pychess.org',
    download_url='http://pychess.org/download/',
    package_dir={'': 'lib'},
    packages=PACKAGES,
    data_files=DATA_FILES,
    scripts=['pychess'],
    options={"build_exe": build_exe_options,
             "bdist_msi": bdist_msi_options},
    executables=executables
)



# S-Chess

from pychess.Utils.const import SCHESS, VARIANTS_OTHER_NONSTANDARD, \
    KNIGHT_PROMOTION, BISHOP_PROMOTION, ROOK_PROMOTION, QUEEN_PROMOTION, \
    HAWK_PROMOTION, ELEPHANT_PROMOTION
from pychess.Utils.Board import Board

SCHESSSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR[heHE] w KQBCDFGkqbcdfg - 0 1"


class SchessBoard(Board):
    variant = SCHESS
    __desc__ = _(
        "S-Chess: https://en.wikipedia.org/wiki/Seirawan_chess")
    name = _("S-Chess")
    cecp_name = "seirawan"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD
    PROMOTIONS = (KNIGHT_PROMOTION, BISHOP_PROMOTION, ROOK_PROMOTION, QUEEN_PROMOTION, HAWK_PROMOTION, ELEPHANT_PROMOTION)

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=SCHESSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



import asyncio
import os
import sys
import pstats
import inspect
import cProfile
import ssl
import tempfile
from timeit import default_timer
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from pychess import MSYS2

# TODO: this hack only needed until we can figure out
# what ssl cert files needed to copied in setup.py
# when cx_freezee bdist_msi command is used
if MSYS2:
    ssl._create_default_https_context = ssl._create_unverified_context


async def download_file_async(url, progressbar=None):
    loop = asyncio.get_event_loop()
    future = loop.run_in_executor(None, download_file, url)
    temp_file = await future
    return temp_file


def download_file(url, progressbar=None):
    temp_file = None
    try:
        if progressbar is not None:
            from gi.repository import GLib
            GLib.idle_add(progressbar.set_text, "Downloading %s ..." % url)
        else:
            print("Downloading %s ..." % url)
        f = urlopen(url)

        temp_file = os.path.join(tempfile.gettempdir(), os.path.basename(url))
        with open(temp_file, "wb") as local_file:
            local_file.write(f.read())

    except HTTPError as e:
        print("HTTP Error:", e.code, url)

    except URLError as e:
        print("URL Error:", e.reason, url)

    return temp_file


def fident(f):
    '''
    Get an identifier for a function or method
    '''
    joinchar = '.'
    if hasattr(f, 'im_class'):
        fparent = f.im_class.__name__
    else:
        joinchar = ':'
        fparent = f.__module__.split('.')[-1]

    # sometimes inspect.getsourcelines() segfaults on windows
    if getattr(sys, 'frozen', False) or sys.platform == "win32":
        lineno = 0
    else:
        lineno = inspect.getsourcelines(f)[1]

    fullname = joinchar.join((fparent, f.__name__))
    return ':'.join((fullname, str(lineno)))


def get_threadname(thread_namer):
    if isinstance(thread_namer, str):
        return thread_namer
    else:
        return fident(thread_namer)


# https://gist.github.com/techtonik/2151727
def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

       `skip` specifies how many levels of stack to skip while getting caller
       name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

       An empty string is returned if skipped levels exceed stack height
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ''
    parentframe = stack[start][0]

    name = []
    module = inspect.getmodule(parentframe)
    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)
    # detect classname
    if 'self' in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals['self'].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename)   # function or a method
    del parentframe
    return ".".join(name)


def profile_me(fn):
    def profiled_fn(*args, **kwargs):
        prof = cProfile.Profile()
        ret = prof.runcall(fn, *args, **kwargs)
        ps = pstats.Stats(prof)
        ps.sort_stats('cumulative')
        ps.print_stats(60)
        return ret
    return profiled_fn


# Python Timer Class - Context Manager for Timing Code Blocks
# Corey Goldberg - 2012
class Timer:
    def __init__(self, text):
        self.text = text
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed_secs = end - self.start
        self.elapsed = self.elapsed_secs * 1000  # millisecs
        print('---- elapsed time: %f ms - %s' % (self.elapsed, self.text))



import platform
import sys


def get_cpu():
    result = {}

    # Operating system
    result['platform'] = sys.platform
    result['windows'] = result['platform'] == 'win32'
    result['linux'] = result['platform'].startswith('linux')
    result['mac'] = result['platform'] == 'darwin'
    result['release'] = platform.release().lower()
    result['binext'] = '.exe' if result['windows'] else ''

    # Number of bits
    result['bitness'] = '64' if platform.machine().endswith('64') or sys.maxsize > 2**32 else '32'

    # Instruction sets
    try:
        info = ''
        with open('/proc/cpuinfo') as f:
            info = f.read()
        result['popcnt'] = 'popcnt' in info
        result['bmi2'] = 'bmi2' in info
    except OSError:
        # Logic not fully true
        guess = (result['bitness'] == '64')
        result['popcnt'] = guess
        result['bmi2'] = guess
    return result



""" Losers Variant"""

from pychess.Utils.const import LOSERSCHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board


class LosersBoard(Board):
    """:Description: The Losers variant is a game where the concept is to get rid of all your
        pieces before you opponent does. On a players turn if a piece can be taken it must be taken
        otherwise a normal chess move can be played
    """
    variant = LOSERSCHESS
    __desc__ = _("FICS losers: http://www.freechess.org/Help/HelpFiles/losers_chess.html")
    name = _("Losers")
    cecp_name = "losers"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD


def testKingOnly(board):
    """ Checks to see if if a winning position has been acheived
    """
    return bin(board.friends[board.color]).count("1") == 1



from gi.repository import Rsvg

from pychess.Utils.const import BLACK, WHITE, KING, QUEEN, BISHOP, KNIGHT, ROOK, PAWN, \
    reprSign, SITTUYINCHESS, HAWK, ELEPHANT, SCHESS
from pychess.System import conf
from pychess.System.prefix import addDataPrefix
from pychess.System.cairoextras import create_cairo_font_face_for_file


piece_ord = {KING: 0, QUEEN: 1, ROOK: 2, BISHOP: 3, KNIGHT: 4, PAWN: 5}
pnames = ('Pawn', 'Knight', 'Bishop', 'Rook', 'Queen', 'King')

size = 800.0


def drawPiece3(piece, context, x, y, psize, allwhite=False, asean=False, variant=None):
    """Rendering pieces using .svg chess figurines"""

    color = WHITE if allwhite else piece.color
    if variant is not None and variant == SITTUYINCHESS:
        image = sittuyin_svg_pieces[color][piece.sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0
    elif asean:
        image = makruk_svg_pieces[color][piece.sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0
    elif all_in_one:
        image = svg_pieces
        w, h = image.props.width / 6, image.props.height / 2
        offset_x = piece_ord[piece.sign] * psize
        offset_y = 0 if color == BLACK else psize
    else:
        if variant is not None and variant == SCHESS:
            image = schess_svg_pieces[color][piece.sign]
        else:
            image = svg_pieces[color][piece.sign]
        w, h = image.props.width, image.props.height
        offset_x = 0
        offset_y = 0

    context.save()

    context.rectangle(x, y, psize, psize)
    context.clip()
    context.translate(x - offset_x, y - offset_y)
    context.scale(1.0 * psize / w, 1.0 * psize / h)

    context.push_group()

    if asean:
        image.render_cairo(context)
    elif all_in_one:
        pieceid = '#%s%s' % ('White' if color == 0 else 'Black',
                             pnames[piece.sign - 1])
        image.render_cairo_sub(context, id=pieceid)
    else:
        image.render_cairo(context)

    context.pop_group_to_source()
    context.paint_with_alpha(piece.opacity)
    context.restore()


def drawPiece4(piece, context, x, y, psize, allwhite=False, asean=False, variant=None):
    """Rendering pieces using .ttf chessfont figurines"""

    if asean:
        drawPiece3(piece, context, x, y, psize, allwhite=allwhite, asean=True)
        return

    color = WHITE if allwhite else piece.color

    context.set_font_face(chess_font_face)
    context.set_font_size(psize)
    context.move_to(x, y + psize)

    context.text_path(piece2char[color][piece.sign])
    close_path = False
    for cmd, points in context.copy_path():
        if cmd == 0:
            context.move_to(*points)
            if close_path:
                context.set_source_rgb(1, 1, 1)
                context.fill_preserve()
                context.set_source_rgb(0, 0, 0)
                close_path = False
        elif cmd == 1:
            context.line_to(*points)
        elif cmd == 2:
            context.curve_to(*points)
        else:
            close_path = True
    context.fill()


surfaceCache = {}

pieces = (PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, HAWK, ELEPHANT)


def get_svg_pieces(svgdir):
    """Load figurines from .svg files"""

    if all_in_one:
        rsvg_handles = Rsvg.Handle.new_from_file(addDataPrefix(
            "pieces/%s/%s.svg" % (svgdir, svgdir)))
    else:
        rsvg_handles = [[None] * 9, [None] * 9]
        for c, color in ((WHITE, 'white'), (BLACK, 'black')):
            for p in pieces:
                if p in (HAWK, ELEPHANT) and svgdir != "merida":
                    continue
                rsvg_handles[c][p] = Rsvg.Handle.new_from_file(addDataPrefix(
                    "pieces/%s/%s%s.svg" % (svgdir, color[0], reprSign[
                        p].lower())))
    return rsvg_handles


def get_chess_font_face(name):
    """Set chess font and char mapping for a chess .ttf"""
    name = name[4:]
    if name in ('alpha', 'berlin', 'cheq'):
        char_map = ('phbrqk', 'ojntwl')
    else:
        char_map = ('pnbrqk', 'omvtwl')

    piece_chars = [[None] * 7, [None] * 7]
    for color in (WHITE, BLACK):
        for piece, char in zip(pieces, char_map[color]):
            piece_chars[color][piece] = char

    face = create_cairo_font_face_for_file(addDataPrefix("pieces/ttf/%s.ttf" %
                                                         name))
    return face, piece_chars


all_in_one = None
drawPiece = None
svg_pieces = None
makruk_svg_pieces = get_svg_pieces("makruk")
sittuyin_svg_pieces = get_svg_pieces("sittuyin")
schess_svg_pieces = get_svg_pieces("merida")
chess_font_face = None
piece2char = None


def set_piece_theme(piece_set):
    global all_in_one
    global drawPiece
    global svg_pieces
    global chess_font_face
    global piece2char

    piece_set = piece_set.lower()
    if piece_set.startswith("ttf-"):
        drawPiece = drawPiece4
        try:
            chess_font_face, piece2char = get_chess_font_face(piece_set)
        except Exception:
            print("Can't create ttf piece set %s" % piece_set)
        all_in_one = False
        drawPiece = drawPiece3
        svg_pieces = get_svg_pieces("merida")
    elif piece_set in ('celtic', 'eyes', 'fantasy', 'fantasy_alt', 'freak',
                       'prmi', 'skulls', 'spatial'):
        all_in_one = True
        drawPiece = drawPiece3
        svg_pieces = get_svg_pieces(piece_set)
    else:
        all_in_one = False
        drawPiece = drawPiece3
        try:
            svg_pieces = get_svg_pieces(piece_set)
        except Exception:
            from pychess.gfx.pychess_pieces import drawPiece2
            drawPiece = drawPiece2


set_piece_theme(conf.get("pieceTheme"))



from gi.repository import GObject

from pychess.ic.VerboseTelnet import ConsoleHandler


class ConsoleManager(GObject.GObject):

    __gsignals__ = {
        'consoleMessage': (GObject.SignalFlags.RUN_FIRST, None, (object,
                                                                 object, )),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.connection.client.lines.consolehandler = ConsoleHandler(
            self.onConsoleMessage)

    def onConsoleMessage(self, lines, ini_lines=None):
        self.emit("consoleMessage", lines, ini_lines)
